@@ -140,5 +140,5 @@ func connect(options xmpp.Options, directTLS bool, anon bool, retryWait int, ret
 		time.Sleep(sleepDuration)
 		slog.Info("connecting to server:", "retry", i)
 	}
-	return client, fmt.Errorf("connect: failed to connect to server: %v", err)
+	return client, fmt.Errorf("connect: failed to connect to server: %w", err)
 }
