@@ -46,11 +46,11 @@ func MarshalJID(input string) (string, error) {
 
 	sep = strings.Index(s, "@")
 
-	switch {
-	case sep == -1:
+	switch sep {
+	case -1:
 		// There is no @ sign, and therefore no localpart.
 		domainpart = s
-	case sep == 0:
+	case 0:
 		// The JID starts with an @ sign (invalid empty localpart)
 		err = fmt.Errorf("invalid jid: %s", input)
 		return input, err
